/*
    Melody maid
    copyright (c) 1998-2004 Kazuki IWAMOTO http://www.maid.org/ iwm@maid.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <gdk/gdkkeysyms.h>
#include "charlist.h"
#include "edit.h"
#include "general.h"
#include "mm_list.h"
#include "misc/fileio.h"


/******************************************************************************
*                                                                             *
* ja:編集ダイアログ関数群                                                     *
*                                                                             *
******************************************************************************/
static GtkWidget *dialog, *button9, *check, *entry0, *entry4, *spin;
static GtkWidget *combo1, *combo2, *combo3, *combo5, *combo7, *combo8;
static GtkWidget *label0, *label1, *label2, *label3;
static GtkWidget *label4, *label5, *label7, *label8;


/* ja:参照でOKボタンが押された */
static void
edit_dialog_clicked_browse_ok (GtkWidget *widget,
                               GtkWidget *dialog_browse)
{
  gchar **file;

  file = g_object_get_data (G_OBJECT (dialog_browse), "user_data");
  *file = g_filename_to_utf8 (gtk_file_selection_get_filename
                (GTK_FILE_SELECTION (dialog_browse)), -1, NULL, NULL, NULL);
  gtk_widget_destroy (dialog_browse);
}


/* ja:参照ボタンが押された */
static void
edit_dialog_clicked_browse (GtkWidget *widget,
                            gpointer   user_data)
{
  gchar *file = NULL;
  GtkWidget *dialog_browse;

  dialog_browse = gtk_file_selection_new (_("Browse"));
  g_signal_connect (G_OBJECT (dialog_browse), "destroy",
                                            G_CALLBACK (gtk_main_quit), NULL);
  g_signal_connect (G_OBJECT (GTK_FILE_SELECTION (dialog_browse)->ok_button),
       "clicked", G_CALLBACK (edit_dialog_clicked_browse_ok), dialog_browse);
  g_signal_connect_swapped
                (G_OBJECT (GTK_FILE_SELECTION (dialog_browse)->cancel_button),
       "clicked", G_CALLBACK (gtk_widget_destroy), GTK_OBJECT (dialog_browse));
  g_object_set_data (G_OBJECT (dialog_browse), "user_data", &file);
  gtk_file_selection_set_filename (GTK_FILE_SELECTION (dialog_browse),
                g_filename_from_utf8 (gtk_entry_get_text (GTK_ENTRY (entry0)),
                                                        -1, NULL, NULL, NULL));
  gtk_file_selection_hide_fileop_buttons (GTK_FILE_SELECTION (dialog_browse));
  gtk_widget_show (dialog_browse);
  gtk_grab_add (dialog_browse);
  gtk_main ();
  if (file)
    {
      gtk_entry_set_text (GTK_ENTRY (entry0), file);
      g_free (file);
    }
}


/* ja:OKボタンが押された */
static void
edit_dialog_clicked_ok (GtkWidget   *widget,
                        MmaidWindow *mmaid)
{
  gboolean f, t, a, l, y, c, r, g, e;
  gchar *file = NULL;
  const gchar *title, *artist, *album;
  const gchar *year, *comment, *genre, *charset;
  gint i, count, track;

  if (entry0)
    {
      gchar *tmp;

      file = g_filename_from_utf8 (gtk_entry_get_text (GTK_ENTRY (entry0)), -1,
                                                            NULL, NULL, NULL);
      tmp = fileio_get_full_path (file);
      g_free (file);
      file = g_filename_to_utf8 (tmp, -1, NULL, NULL, NULL);
      g_free (tmp);
      if (!file)
        return;
      f = gtk_label_get_use_markup (GTK_LABEL (label0));
    }
  else
    {
      f = FALSE;
    }
  title   = gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (combo1)->entry));
  artist  = gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (combo2)->entry));
  album   = gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (combo3)->entry));
  year    = gtk_entry_get_text (GTK_ENTRY (entry4));
  comment = gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (combo5)->entry));
  track   = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (check))
              ? gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (spin)) : -1;
  genre   = gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (combo7)->entry));
  charset = gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (combo8)->entry));
  t = gtk_label_get_use_markup (GTK_LABEL (label1));
  a = gtk_label_get_use_markup (GTK_LABEL (label2));
  l = gtk_label_get_use_markup (GTK_LABEL (label3));
  y = gtk_label_get_use_markup (GTK_LABEL (label4));
  c = gtk_label_get_use_markup (GTK_LABEL (label5));
  r = gtk_label_get_use_markup
                            (GTK_LABEL (gtk_bin_get_child (GTK_BIN (check))));
  g = gtk_label_get_use_markup (GTK_LABEL (label7));
  e = gtk_label_get_use_markup (GTK_LABEL (label8));

  count = mm_list_length (MM_LIST (mmaid->list));
  for (i = 0; i < count; i++)
    if (mm_list_is_selected (MM_LIST (mmaid->list), i))
      {
          if (f)
            mm_list_change (MM_LIST (mmaid->list), i, MM_LIST_FILE,
                                                        g_strdup (file), -1);
          if (t)
            mm_list_change (MM_LIST (mmaid->list), i, MM_LIST_TITLE,
                          title[0] != '\0' ? g_strdup (title)   : NULL, -1);
          if (a)
            mm_list_change (MM_LIST (mmaid->list), i, MM_LIST_ARTIST,
                         artist[0] != '\0' ? g_strdup (artist)  : NULL, -1);
          if (l)
            mm_list_change (MM_LIST (mmaid->list), i, MM_LIST_ALBUM,
                          album[0] != '\0' ? g_strdup (album)   : NULL, -1);
          if (y)
            mm_list_change (MM_LIST (mmaid->list), i, MM_LIST_YEAR,
                           year[0] != '\0' ? g_strdup (year)    : NULL, -1);
          if (c)
            mm_list_change (MM_LIST (mmaid->list), i, MM_LIST_COMMENT,
                        comment[0] != '\0' ? g_strdup (comment) : NULL, -1);
          if (r)
            mm_list_change (MM_LIST (mmaid->list), i, MM_LIST_TRACK,
                                                                    track, -1);
          if (g)
            mm_list_change (MM_LIST (mmaid->list), i, MM_LIST_GENRE,
                          genre[0] != '\0' ? g_strdup (genre)   : NULL, -1);
          if (e)
            mm_list_change (MM_LIST (mmaid->list), i, MM_LIST_CHARSET,
                        charset[0] != '\0' ? g_strdup (charset) : NULL, -1);
      }
  g_free (file);
  if (e && charset[0] != '\0')
    charlist_renewal_all (charset);
  mmaid_change_edit (mmaid, TRUE);
  gtk_widget_destroy (dialog);
}


/* ja:コンボボックス/エントリー/スピンボタンでリターンが押された */
static void
edit_dialog_activate (GtkWidget *widget,
                      gpointer   user_data)
{
  if (GTK_WIDGET_IS_SENSITIVE (button9))
    edit_dialog_clicked_ok (widget, user_data);
}


/* ja:チェックボタンが変更された */
#define edit_dialog_toggled edit_dialog_changed


/* ja:コンボボックス/エントリー/スピンボタンが変更された */
static void
edit_dialog_changed (GtkWidget *widget,
                     GtkWidget *label)
{
  gboolean result = TRUE;
  const gchar *charset;
  GtkWidget *list;

  list = g_object_get_data (G_OBJECT (widget), "user_data");
  if (!gtk_label_get_use_markup (GTK_LABEL (label)))
    {
      gchar *str;

      str = g_strconcat ("<b><i>", gtk_label_get_label (GTK_LABEL (label)),
                                                            "</i></b>*", NULL);
      gtk_label_set_markup_with_mnemonic (GTK_LABEL (label), str);
      g_free (str);
    }
  charset = gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (combo8)->entry));
  if (entry0)
    {
      const gchar *file;
      gint i;

      file = gtk_entry_get_text (GTK_ENTRY (entry0));
      for (i = mm_list_length (MM_LIST (list)) - 1; i >= 0; i--)
        if (mm_list_is_selected (MM_LIST (list), i))
          break;
      result = mm_list_is_valid_file (MM_LIST (list), i, file);
    }
  if (charset[0] != '\0' && !charlist_is_valid (charset))
    result = FALSE;
  gtk_widget_set_sensitive (button9, result);
}


/* ja:ESCが押された */
static gboolean
edit_dialog_key_press (GtkWidget   *widget,
                       GdkEventKey *event,
                       gpointer     user_data)
{
  if (event->keyval == GDK_Escape)
    gtk_widget_destroy (widget);
  return FALSE;
}


void
edit_dialog (MmaidWindow *mmaid)
{
  gchar **charset;
  gint i, c, s = 0, count;
  GList *glist1 = NULL, *glist2 = NULL, *glist3 = NULL;
  GList *glist5 = NULL, *glist7 = NULL, *glist8 = NULL;
  GtkWidget *button0, *button10, *table, *hbox, *vbox;
  const ID3Tag *id3tag = NULL;

  /* ja:メインウインドウ */
  dialog = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (dialog), _("Edit"));
  g_signal_connect_after (G_OBJECT (dialog), "key-press-event",
                                    G_CALLBACK (edit_dialog_key_press), NULL);
  g_signal_connect (G_OBJECT (dialog), "destroy",
                                            G_CALLBACK (gtk_main_quit), NULL);
  /* ja:ボタン */
  button9  = gtk_button_new_with_label (_("OK"));
  button10 = gtk_button_new_with_label (_("Cancel"));
  g_signal_connect (G_OBJECT (button9), "clicked",
                        G_CALLBACK (edit_dialog_clicked_ok), mmaid);
  g_signal_connect_swapped (G_OBJECT (button10), "clicked",
                        G_CALLBACK (gtk_widget_destroy), G_OBJECT (dialog));
  GTK_WIDGET_SET_FLAGS (button9,  GTK_CAN_DEFAULT);
  GTK_WIDGET_SET_FLAGS (button10, GTK_CAN_DEFAULT);
  /* ja:ラベル */
  label1 = gtk_label_new_with_mnemonic (_("_Title"));
  label2 = gtk_label_new_with_mnemonic (_("_Artist"));
  label3 = gtk_label_new_with_mnemonic (_("A_lbum"));
  label4 = gtk_label_new_with_mnemonic (_("_Year"));
  label5 = gtk_label_new_with_mnemonic (_("_Comment"));
  label7 = gtk_label_new_with_mnemonic (_("_Genre"));
  label8 = gtk_label_new_with_mnemonic (_("Character _Encoding"));
  /* ja:コンボボックス(準備) */
  count = mm_list_length (MM_LIST (mmaid->list));
  for (i = 0; i < count; i++)
    {
      id3tag = mm_list_get_nth (MM_LIST (mmaid->list), i);
      if (id3tag->title   && g_list_find_custom (glist1, id3tag->title,
                                                (GCompareFunc)strcmp) == NULL)
        glist1 = g_list_insert_sorted (glist1, g_strdup (id3tag->title),
                                                (GCompareFunc)strcmp);
      if (id3tag->artist  && g_list_find_custom (glist2, id3tag->artist,
                                                (GCompareFunc)strcmp) == NULL)
        glist2 = g_list_insert_sorted (glist2, g_strdup (id3tag->artist),
                                                (GCompareFunc)strcmp);
      if (id3tag->album   && g_list_find_custom (glist3, id3tag->album,
                                                (GCompareFunc)strcmp) == NULL)
        glist3 = g_list_insert_sorted (glist3, g_strdup (id3tag->album),
                                                (GCompareFunc)strcmp);
      if (id3tag->comment && g_list_find_custom (glist5, id3tag->comment,
                                                (GCompareFunc)strcmp) == NULL)
        glist5 = g_list_insert_sorted (glist5, g_strdup (id3tag->comment),
                                                (GCompareFunc)strcmp);
    }
  for (i = 0; i < 256; i++)
    if (genre[0][i] && g_list_find_custom (glist7, genre[0][i],
                                                (GCompareFunc)strcmp) == NULL)
      glist7 = g_list_append (glist7, g_strdup (genre[0][i]));
  charset = g_strsplit (charset_list, ",", G_MAXINT);
  for (i = 0; charset[i]; i++)
    glist8 = g_list_insert_sorted (glist8, g_strdup (charset[i]),
                                                        (GCompareFunc)strcmp);
  g_strfreev (charset);
  /* ja:選択の数を数える */
  for (i = 0; i < count; i++)
    if (mm_list_is_selected (MM_LIST (mmaid->list), i))
      {
        s++;
        if (s > 1)
          break;
        id3tag = mm_list_get_nth (MM_LIST (mmaid->list), i);
      }
  if (s <= 0)
    return;
  /* ja:コンボボックス */
  combo1 = gtk_combo_new ();
  combo2 = gtk_combo_new ();
  combo3 = gtk_combo_new ();
  combo5 = gtk_combo_new ();
  combo7 = gtk_combo_new ();
  combo8 = gtk_combo_new ();
  if (glist1)
    gtk_combo_set_popdown_strings (GTK_COMBO (combo1), glist1);
  if (glist2)
    gtk_combo_set_popdown_strings (GTK_COMBO (combo2), glist2);
  if (glist3)
    gtk_combo_set_popdown_strings (GTK_COMBO (combo3), glist3);
  if (glist5)
    gtk_combo_set_popdown_strings (GTK_COMBO (combo5), glist5);
  if (glist7)
    gtk_combo_set_popdown_strings (GTK_COMBO (combo7), glist7);
  if (glist8)
    gtk_combo_set_popdown_strings (GTK_COMBO (combo8), glist8);
  gtk_combo_set_value_in_list (GTK_COMBO (combo1), FALSE, FALSE);
  gtk_combo_set_value_in_list (GTK_COMBO (combo2), FALSE, FALSE);
  gtk_combo_set_value_in_list (GTK_COMBO (combo3), FALSE, FALSE);
  gtk_combo_set_value_in_list (GTK_COMBO (combo5), FALSE, FALSE);
  gtk_combo_set_value_in_list (GTK_COMBO (combo7), FALSE, FALSE);
  gtk_combo_set_value_in_list (GTK_COMBO (combo8), FALSE, FALSE);
  gtk_combo_set_case_sensitive (GTK_COMBO (combo1), TRUE);
  gtk_combo_set_case_sensitive (GTK_COMBO (combo2), TRUE);
  gtk_combo_set_case_sensitive (GTK_COMBO (combo3), TRUE);
  gtk_combo_set_case_sensitive (GTK_COMBO (combo5), TRUE);
  gtk_combo_set_case_sensitive (GTK_COMBO (combo7), TRUE);
  gtk_combo_set_case_sensitive (GTK_COMBO (combo8), TRUE);
  if (s == 1)
    {
      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO (combo1)->entry),
                                    id3tag->title ? id3tag->title : "");
      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO (combo2)->entry),
                                    id3tag->artist ? id3tag->artist : "");
      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO (combo3)->entry),
                                    id3tag->album ? id3tag->album : "");
      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO (combo5)->entry),
                                    id3tag->comment ? id3tag->comment : "");
      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO (combo7)->entry),
                                    id3tag->genre ? id3tag->genre : "");
      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO (combo8)->entry),
                                    id3tag->charset ? id3tag->charset : "");
    }
  else
    {
      gboolean t = TRUE, a = TRUE, l = TRUE, c = TRUE, g = TRUE, e = TRUE;

      for (i = 0; i < count && (t || a || l || c || g || e); i++)
        if (mm_list_is_selected (MM_LIST (mmaid->list), i))
          {
            const ID3Tag *id3tag0;

            id3tag0 = mm_list_get_nth (MM_LIST (mmaid->list), i);
            t = t && id3tag0->title
                        && g_strcmp (id3tag->title,   id3tag0->title) == 0;
            a = a && id3tag0->artist
                        && g_strcmp (id3tag->artist,  id3tag0->artist) == 0;
            l = l && id3tag0->album
                        && g_strcmp (id3tag->album,   id3tag0->album) == 0;
            c = c && id3tag0->comment
                        && g_strcmp (id3tag->comment, id3tag0->comment) == 0;
            g = g && id3tag0->genre
                        && g_strcmp (id3tag->genre,   id3tag0->genre) == 0;
            e = e && id3tag0->charset
                        && g_strcmp (id3tag->charset, id3tag0->charset) == 0;
          }
      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO (combo1)->entry),
                                                    t ? id3tag->title   : "");
      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO (combo2)->entry),
                                                    a ? id3tag->artist  : "");
      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO (combo3)->entry),
                                                    l ? id3tag->album   : "");
      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO (combo5)->entry),
                                                    c ? id3tag->comment : "");
      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO (combo7)->entry),
                                                    g ? id3tag->genre   : "");
      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO (combo8)->entry),
                                                    e ? id3tag->charset : "");
    }
  g_signal_connect (G_OBJECT (GTK_COMBO (combo1)->entry), "activate",
                                    G_CALLBACK (edit_dialog_activate), mmaid);
  g_signal_connect (G_OBJECT (GTK_COMBO (combo2)->entry), "activate",
                                    G_CALLBACK (edit_dialog_activate), mmaid);
  g_signal_connect (G_OBJECT (GTK_COMBO (combo3)->entry), "activate",
                                    G_CALLBACK (edit_dialog_activate), mmaid);
  g_signal_connect (G_OBJECT (GTK_COMBO (combo5)->entry), "activate",
                                    G_CALLBACK (edit_dialog_activate), mmaid);
  g_signal_connect (G_OBJECT (GTK_COMBO (combo7)->entry), "activate",
                                    G_CALLBACK (edit_dialog_activate), mmaid);
  g_signal_connect (G_OBJECT (GTK_COMBO (combo8)->entry), "activate",
                                    G_CALLBACK (edit_dialog_activate), mmaid);
  g_signal_connect (G_OBJECT (GTK_COMBO (combo1)->entry), "changed",
                                    G_CALLBACK (edit_dialog_changed), label1);
  g_signal_connect (G_OBJECT (GTK_COMBO (combo2)->entry), "changed",
                                    G_CALLBACK (edit_dialog_changed), label2);
  g_signal_connect (G_OBJECT (GTK_COMBO (combo3)->entry), "changed",
                                    G_CALLBACK (edit_dialog_changed), label3);
  g_signal_connect (G_OBJECT (GTK_COMBO (combo5)->entry), "changed",
                                    G_CALLBACK (edit_dialog_changed), label5);
  g_signal_connect (G_OBJECT (GTK_COMBO (combo7)->entry), "changed",
                                    G_CALLBACK (edit_dialog_changed), label7);
  g_signal_connect (G_OBJECT (GTK_COMBO (combo8)->entry), "changed",
                                    G_CALLBACK (edit_dialog_changed), label8);
  for (i = g_list_length (glist1) - 1; i >= 0; i--)
    g_free (g_list_nth_data (glist1, i));
  for (i = g_list_length (glist2) - 1; i >= 0; i--)
    g_free (g_list_nth_data (glist2, i));
  for (i = g_list_length (glist3) - 1; i >= 0; i--)
    g_free (g_list_nth_data (glist3, i));
  for (i = g_list_length (glist5) - 1; i >= 0; i--)
    g_free (g_list_nth_data (glist5, i));
  for (i = g_list_length (glist7) - 1; i >= 0; i--)
    g_free (g_list_nth_data (glist7, i));
  for (i = g_list_length (glist8) - 1; i >= 0; i--)
    g_free (g_list_nth_data (glist8, i));
  g_list_free (glist1);
  g_list_free (glist2);
  g_list_free (glist3);
  g_list_free (glist5);
  g_list_free (glist7);
  g_list_free (glist8);
  /* ja:エントリー */
  entry4 = gtk_entry_new ();
  if (s == 1 && id3tag->year)
    gtk_entry_set_text (GTK_ENTRY (entry4), id3tag->year);
  g_signal_connect (G_OBJECT (entry4), "activate",
                                    G_CALLBACK (edit_dialog_activate), mmaid);
  g_signal_connect (G_OBJECT (entry4), "changed",
                                    G_CALLBACK (edit_dialog_changed), label4);
  /* ja:チェックボタン */
  check = gtk_check_button_new_with_mnemonic (_("T_rack"));
  if (s == 1)
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (check),
                                0 <= id3tag->track && id3tag->track <= 255);
  g_signal_connect (G_OBJECT (check), "toggled",
        G_CALLBACK (edit_dialog_toggled), gtk_bin_get_child (GTK_BIN (check)));
  /* ja:スピンボタン */
  spin = gtk_spin_button_new (GTK_ADJUSTMENT (gtk_adjustment_new
     (s == 1 && 0 <= id3tag->track && id3tag->track <= 255 ? id3tag->track : 0,
                                                    0, 255, 1, 5, 0)), 0, 0);
  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (spin), TRUE);
  g_signal_connect (G_OBJECT (spin), "activate",
                                    G_CALLBACK (edit_dialog_activate), mmaid);
  g_signal_connect (G_OBJECT (spin), "changed",
        G_CALLBACK (edit_dialog_changed), gtk_bin_get_child (GTK_BIN (check)));
  /* ja:ラベル */
  gtk_label_set_mnemonic_widget (GTK_LABEL(label1), GTK_COMBO (combo1)->entry);
  gtk_label_set_mnemonic_widget (GTK_LABEL(label2), GTK_COMBO (combo2)->entry);
  gtk_label_set_mnemonic_widget (GTK_LABEL(label3), GTK_COMBO (combo3)->entry);
  gtk_label_set_mnemonic_widget (GTK_LABEL(label4), entry4);
  gtk_label_set_mnemonic_widget (GTK_LABEL(label5), GTK_COMBO (combo5)->entry);
  gtk_label_set_mnemonic_widget (GTK_LABEL(label7), GTK_COMBO (combo7)->entry);
  gtk_label_set_mnemonic_widget (GTK_LABEL(label8), GTK_COMBO (combo8)->entry);
  /* ja:ファイル */
  if (s == 1)
    {
      /* ja:ボタン */
      button0 = gtk_button_new_with_mnemonic (_("_Browse..."));
      g_signal_connect (G_OBJECT (button0), "clicked",
                        G_CALLBACK (edit_dialog_clicked_browse), NULL);
      GTK_WIDGET_SET_FLAGS (button0,  GTK_CAN_DEFAULT);
      /* ja:ラベル */
      label0 = gtk_label_new_with_mnemonic (_("_File"));
      /* ja:エントリー */
      entry0 = gtk_entry_new ();
      gtk_entry_set_text (GTK_ENTRY (entry0), id3tag->file);
      g_signal_connect (G_OBJECT (entry0), "activate",
                                    G_CALLBACK (edit_dialog_activate), mmaid);
      g_signal_connect (G_OBJECT (entry0), "changed",
                                    G_CALLBACK (edit_dialog_changed), label0);
      /* ja:ラベル */
      gtk_label_set_mnemonic_widget (GTK_LABEL(label0), entry0);
    }
  else
    {
      button0 = entry0 = label0 = NULL;
    }
  /* ja:リスト */
  if (entry0)
    g_object_set_data (G_OBJECT (entry0), "user_data", mmaid->list);
  g_object_set_data (G_OBJECT (GTK_COMBO (combo1)->entry), "user_data",
                                                                mmaid->list);
  g_object_set_data (G_OBJECT (GTK_COMBO (combo2)->entry), "user_data",
                                                                mmaid->list);
  g_object_set_data (G_OBJECT (GTK_COMBO (combo3)->entry), "user_data",
                                                                mmaid->list);
  g_object_set_data (G_OBJECT (entry4), "user_data", mmaid->list);
  g_object_set_data (G_OBJECT (GTK_COMBO (combo5)->entry), "user_data",
                                                                mmaid->list);
  g_object_set_data (G_OBJECT (spin), "user_data", mmaid->list);
  g_object_set_data (G_OBJECT (GTK_COMBO (combo7)->entry), "user_data",
                                                                mmaid->list);
  g_object_set_data (G_OBJECT (GTK_COMBO (combo8)->entry), "user_data",
                                                                mmaid->list);
  g_object_set_data (G_OBJECT (check), "user_data", mmaid->list);
  /* ja:フレームとボックス */
  table = gtk_table_new (s == 1 ? 7 : 6, 4, FALSE);
  gtk_table_set_row_spacings (GTK_TABLE (table), SPACING);
  gtk_table_set_col_spacings (GTK_TABLE (table), SPACING);
  i = 0;
  if (s == 1)
    {
      hbox = gtk_hbox_new (FALSE, 0);
      gtk_box_pack_end (GTK_BOX (hbox), label0, FALSE, FALSE, 0);
      gtk_table_attach (GTK_TABLE (table), hbox, 0, 1, i, i + 1,
                                                    GTK_FILL, GTK_FILL, 0, 0);
      hbox = gtk_hbox_new (FALSE, SPACING);
      gtk_box_pack_start (GTK_BOX (hbox), entry0, TRUE, TRUE, 0);
      gtk_box_pack_start (GTK_BOX (hbox), button0, FALSE, FALSE, 0);
      gtk_table_attach_defaults (GTK_TABLE (table), hbox,   1, 4, i, i + 1);
      i++;
    }
  hbox = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_end (GTK_BOX (hbox), label1, FALSE, FALSE, 0);
  gtk_table_attach (GTK_TABLE (table), hbox, 0, 1, i, i + 1,
                                                    GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach_defaults (GTK_TABLE (table), combo1, 1, 4, i, i + 1);
  i++;
  hbox = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_end (GTK_BOX (hbox), label2, FALSE, FALSE, 0);
  gtk_table_attach (GTK_TABLE (table), hbox, 0, 1, i, i + 1,
                                                    GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach_defaults (GTK_TABLE (table), combo2, 1, 4, i, i + 1);
  i++;
  hbox = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_end (GTK_BOX (hbox), label3, FALSE, FALSE, 0);
  gtk_table_attach (GTK_TABLE (table), hbox, 0, 1, i, i + 1,
                                                    GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach_defaults (GTK_TABLE (table), combo3, 1, 4, i, i + 1);
  i++;
  hbox = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_end (GTK_BOX (hbox), label5, FALSE, FALSE, 0);
  gtk_table_attach (GTK_TABLE (table), hbox, 0, 1, i, i + 1,
                                                    GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach_defaults (GTK_TABLE (table), combo5, 1, 4, i, i + 1);
  i++;
  hbox = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_end (GTK_BOX (hbox), label4, FALSE, FALSE, 0);
  gtk_table_attach (GTK_TABLE (table), hbox, 0, 1, i, i + 1,
                                                    GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach_defaults (GTK_TABLE (table), entry4, 1, 2, i, i + 1);
  hbox = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_end (GTK_BOX (hbox), check, FALSE, FALSE, 0);
  gtk_table_attach (GTK_TABLE (table), hbox, 2, 3, i, i + 1,
                                                    GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach_defaults (GTK_TABLE (table), spin,   3, 4, i, i + 1);
  i++;
  hbox = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_end (GTK_BOX (hbox), label7, FALSE, FALSE, 0);
  gtk_table_attach (GTK_TABLE (table), hbox, 0, 1, i, i + 1,
                                                    GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach_defaults (GTK_TABLE (table), combo7, 1, 2, i, i + 1);
  hbox = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_end (GTK_BOX (hbox), label8, FALSE, FALSE, 0);
  gtk_table_attach (GTK_TABLE (table), hbox, 2, 3, i, i + 1,
                                                    GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach_defaults (GTK_TABLE (table), combo8, 3, 4, i, i + 1);
  vbox = gtk_vbox_new (FALSE, SPACING);
  gtk_container_set_border_width (GTK_CONTAINER (vbox), SPACING);
  gtk_box_pack_start (GTK_BOX (vbox), table, TRUE, TRUE, 0);
  hbox = gtk_hbox_new (FALSE, SPACING);
  gtk_box_pack_end (GTK_BOX (hbox), button10, FALSE, FALSE, 0);
  gtk_box_pack_end (GTK_BOX (hbox), button9, FALSE, FALSE, 0);
  gtk_box_pack_end (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
  gtk_container_add (GTK_CONTAINER (dialog), vbox);

  /* ja:表示 */
  gtk_widget_set_sensitive (button9, FALSE);
  mm_list_get_cursor (MM_LIST (mmaid->list), NULL, &c);
  switch (c)
    {
      case MM_LIST_FILE:
        gtk_widget_grab_focus (entry0 ? entry0 : button10); break;
      case MM_LIST_TITLE:
        gtk_widget_grab_focus (GTK_COMBO (combo1)->entry);  break;
      case MM_LIST_ARTIST:
        gtk_widget_grab_focus (GTK_COMBO (combo2)->entry);  break;
      case MM_LIST_ALBUM:
        gtk_widget_grab_focus (GTK_COMBO (combo3)->entry);  break;
      case MM_LIST_YEAR:
        gtk_widget_grab_focus (entry4);                     break;
      case MM_LIST_COMMENT:
        gtk_widget_grab_focus (GTK_COMBO (combo5)->entry);  break;
      case MM_LIST_TRACK:
        gtk_widget_grab_focus (check);                      break;
      case MM_LIST_GENRE:
        gtk_widget_grab_focus (GTK_COMBO (combo7)->entry);  break;
      case MM_LIST_CHARSET:
        gtk_widget_grab_focus (GTK_COMBO (combo8)->entry);  break;
      default:
        gtk_widget_grab_focus (button10);
    }

  gtk_window_set_policy (GTK_WINDOW (dialog), FALSE, TRUE, FALSE);
  gtk_grab_add (dialog);
  gtk_widget_show_all (dialog);
  gtk_main ();
}
